"""Models and schemas for Ert parameters."""

from .distributions import (
    ConstDistribution,
    DerrfDistribution,
    DistributionMetadata,
    DUnifDistribution,
    ErrfDistribution,
    LogNormalDistribution,
    LogUnifDistribution,
    NormalDistribution,
    RawDistribution,
    TriangularDistribution,
    TruncatedNormalDistribution,
    UniformDistribution,
)
from .table_schema import ParameterColumn, ParameterTableSchema

__all__ = [
    "DistributionMetadata",
    "ConstDistribution",
    "DerrfDistribution",
    "DistributionMetadata",
    "DUnifDistribution",
    "ErrfDistribution",
    "LogNormalDistribution",
    "LogUnifDistribution",
    "NormalDistribution",
    "RawDistribution",
    "TriangularDistribution",
    "TruncatedNormalDistribution",
    "UniformDistribution",
    "ParameterColumn",
    "ParameterTableSchema",
]



from typing import Literal

from pydantic import BaseModel

from .distributions import DistributionMetadata


class ParameterColumn(BaseModel):
    """Metadata for a parameter column."""

    type: Literal["float64", "int64"]
    metadata: DistributionMetadata


class ParameterTableSchema(BaseModel):
    """
    Schema for PyArrow parameters exported with Ert parameters.

    The table always has a 'realization' column (int64) as the first column.
    """

    parameters: dict[str, ParameterColumn]



"""Schemas for Ert parameter distributions.

These schemas are used for Parquet column metadata in Ert parameter tables."""

import json
from typing import Annotated, Literal, Self

from pydantic import BaseModel, Field


class BaseDistribution(BaseModel):
    """Base class for all distribution metadata.

    These models are attached as column metadata for the exported Ert parameters
    Parquet table."""

    model_config = {"extra": "forbid"}

    group: str
    input_source: Literal["sampled", "design_matrix"]

    def to_pa_metadata(self) -> dict[bytes, bytes]:
        """Convert the model to a PyArrow-compatible metadata dictionary."""
        return {
            k.encode("utf-8"): json.dumps(v).encode("utf-8")
            for k, v in self.model_dump().items()
        }

    @classmethod
    def from_pa_metadata(cls, metadata: dict[bytes, bytes]) -> Self:
        """Create an instance of the class from PyArrow metadata."""
        str_metadata = {
            k.decode("utf-8"): json.loads(v.decode("utf-8"))
            for k, v in metadata.items()
        }
        return cls.model_validate(str_metadata)


class UniformDistribution(BaseDistribution):
    """Metadata values for a uniform distribution."""

    distribution: Literal["uniform"] = "uniform"
    min: float
    max: float


class LogUnifDistribution(BaseDistribution):
    """Metadata values for a log uniform distribution."""

    distribution: Literal["logunif"] = "logunif"
    min: float
    max: float


class NormalDistribution(BaseDistribution):
    """Metadata values for a normal distribution."""

    distribution: Literal["normal"] = "normal"
    mean: float
    std: float


class LogNormalDistribution(BaseDistribution):
    """Metadata values for a log normal distribution."""

    distribution: Literal["lognormal"] = "lognormal"
    mean: float
    std: float


class TruncatedNormalDistribution(BaseDistribution):
    """Metadata values for a truncated normal distribution."""

    distribution: Literal["truncated_normal"] = "truncated_normal"
    min: float
    max: float
    mean: float
    std: float


class RawDistribution(BaseDistribution):
    """Metadata values for a raw distribution.

    This "distribution" is used for design matrix parameters.
    """

    distribution: Literal["raw"] = "raw"


class ConstDistribution(BaseDistribution):
    """Metadata values for a const distribution."""

    distribution: Literal["const"] = "const"
    value: float


class DUnifDistribution(BaseDistribution):
    """Metadata values for a discrete uniform distribution."""

    distribution: Literal["dunif"] = "dunif"
    min: float
    max: float
    steps: int


class TriangularDistribution(BaseDistribution):
    """Metadata values for a triangular distribution."""

    distribution: Literal["triangular"] = "triangular"
    min: float
    max: float
    mode: float


class ErrfDistribution(BaseDistribution):
    """Metadata values for a Errf (error function) distribution."""

    distribution: Literal["errf"] = "errf"
    min: float
    max: float
    skewness: float
    width: float


class DerrfDistribution(BaseDistribution):
    """Metadata values for a Derrf (discrete error function) distribution."""

    distribution: Literal["derrf"] = "derrf"
    min: float
    max: float
    skewness: float
    width: float
    steps: float


DistributionMetadata = Annotated[
    UniformDistribution
    | LogUnifDistribution
    | LogNormalDistribution
    | NormalDistribution
    | TruncatedNormalDistribution
    | RawDistribution
    | ConstDistribution
    | DUnifDistribution
    | TriangularDistribution
    | ErrfDistribution
    | DerrfDistribution,
    Field(discriminator="distribution"),
]



import pyarrow as pa
import pytest

from fmu.datamodels.parameters.distributions import (
    ConstDistribution,
    DerrfDistribution,
    DistributionMetadata,
    DUnifDistribution,
    ErrfDistribution,
    LogNormalDistribution,
    LogUnifDistribution,
    NormalDistribution,
    RawDistribution,
    TriangularDistribution,
    TruncatedNormalDistribution,
    UniformDistribution,
)


def test_to_pa_metadata_serialization() -> None:
    """Dumping Parquet schemas serializes to dict[bytes, bytes].

    Uses DUnif because it has both float and int."""
    dist = DUnifDistribution(
        min=0.0, max=1.0, steps=100, group="GLOBVAR", input_source="sampled"
    )
    assert dist.to_pa_metadata() == {
        b"group": b'"GLOBVAR"',
        b"input_source": b'"sampled"',
        b"distribution": b'"dunif"',
        b"min": b"0.0",
        b"max": b"1.0",
        b"steps": b"100",
    }

    field = pa.field("PARAM1", pa.float64(), metadata=dist.to_pa_metadata())
    assert field.metadata == {
        b"group": b'"GLOBVAR"',
        b"input_source": b'"sampled"',
        b"distribution": b'"dunif"',
        b"min": b"0.0",
        b"max": b"1.0",
        b"steps": b"100",
    }


@pytest.mark.parametrize(
    "distribution_class, value_dict",
    [
        (UniformDistribution, {"min": 0.0, "max": 1.0}),
        (LogUnifDistribution, {"min": 0.0, "max": 1.0}),
        (NormalDistribution, {"mean": 0.2, "std": 0.1}),
        (
            TruncatedNormalDistribution,
            {"min": 0.0, "max": 1.0, "mean": 0.2, "std": 0.1},
        ),
        (LogNormalDistribution, {"mean": 0.2, "std": 0.1}),
        (RawDistribution, {}),
        (ConstDistribution, {"value": 2.0}),
        (TriangularDistribution, {"min": 0.0, "max": 3.0, "mode": 2.0}),
        (ErrfDistribution, {"min": 0.0, "max": 3.0, "skewness": 2.0, "width": 10.0}),
        (
            DerrfDistribution,
            {"min": 0.0, "max": 3.0, "skewness": 2.0, "width": 10.0, "steps": 1000.0},
        ),
        (DUnifDistribution, {"min": 0.0, "max": 1.0, "steps": 100}),
    ],
)
def test_from_pa_metadata_roundtrip(
    distribution_class: DistributionMetadata, value_dict: dict[str, int | float]
) -> None:
    """Roundtrip serialization-deserialization works for all distributions."""
    input_source = (
        "design_matrix" if type(distribution_class) is RawDistribution else "sampled"
    )
    dist = distribution_class.model_validate(
        {"group": "GLOBVAR", "input_source": input_source, **value_dict}
    )
    field = pa.field("PARAM1", pa.float64(), metadata=dist.to_pa_metadata())
    assert field.metadata is not None
    from_dist = distribution_class.from_pa_metadata(field.metadata)
    assert dist == from_dist
